#include "formule.hpp"
#include "parser.hpp"
#include "adder.hpp"
#include "resolution.hpp"
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <string>

using namespace std;

void help() {
  cout << "Usage:" << endl;
  cout << " ./verifier --help" << endl;
  cout << "    Affiche l'aide" << endl;
  cout << " ./verifier [ --use-minisat ] [ fichier ]" << endl;
  cout << "    Vérifie si une formule est valide. " << endl
       << "    Si fichier est indiqué, la formule est lue dedans." << endl
       << "    Sinon, elle est lue sur l'entrée standard " << endl;
  cout << " ./verifier --check-adder [ --use-minisat ] [ fichier ]" << endl;
  cout << "    Vérifie si un additionneur est correct. " << endl
       << "    Si fichier est indiqué, l'additionneur est lu dedans." << endl
       << "    Sinon, il est lu sur l'entrée standard " << endl;
  cout << endl
       << " --use-minisat utilise minisat au lieu du solveur interne" << endl
       << "               (suppose que la commande minisat soit disponible" << endl
       << "                dans le PATH unix)" << endl;
}

bool minisat_check(map<string,int> & corresp, cnf_t cnf) {
  string filename = "dimacs.tmp";
  ofstream out(filename);
  writeDimacs(out,corresp,cnf);
  out.close();
  int ret = system("minisat -verb=0 dimacs.tmp | grep -q '^SATISFIABLE$'");
  return ret != 0;
}

int main(int argc, char** argv) {
  string file = "";
  bool check_adder = false;
  string adder_opt = "--check-adder";
  bool use_minisat = false;
  string minisat_opt = "--use-minisat";
  string help_opt = "--help";
  formule to_check;
  for(int i = 1; i < argc; ++i) {
    if (string(argv[i]).substr(0,2) == "--") {
      if (string(argv[i]) == help_opt) {
        help();
        return EXIT_SUCCESS;
      } else if (string(argv[i]) == adder_opt) {
        check_adder = true;
      } else if (string(argv[i]) == minisat_opt) {
        use_minisat = true;
      } else {
        cerr << "Option inconnue: " << string(argv[i]) << endl;
        help();
        return EXIT_FAILURE;
      }
    } else {
      file = string(argv[i]);
      cerr << "use file " << file << endl;
    }
  }
  if (check_adder) {
    vector<formule> adder_outputs;
    if ("" == file) {
      cerr << "reading from stdin" << endl;
      adder_outputs = lit_formules(cin);
    } else {
      cerr << "reading from " << file << endl;
      ifstream input(file);
      adder_outputs = lit_formules(input);
      input.close();
    }
    to_check = formule_verifier_additionneur(adder_outputs);
  } else {
    if ("" == file) {
      cerr << "reading from stdin" << endl;
      to_check = lit_formule(cin);
    }
    else {
      cerr << "reading from " << file << endl;
      ifstream input(file);
      to_check = lit_formule(input);
      input.close();
    }
  }
  
  if (to_check == NULL) {
    cerr << "to check is null" << endl;
    return EXIT_FAILURE;
  }
  
  to_check = non(to_check);
  cout << "to_check ready" << endl;
  cout << to_check << endl;
  map<string, var_t> cor;
  numerote(to_check,cor);
  cout << "numerote done" << endl;
  cnf_t cnf;
  tseitin(to_check,cor,cnf);
  cout << "tseitin done" << endl;
  bool neg_unsat;
  if (use_minisat) {
    neg_unsat = minisat_check(cor,cnf);
  } else {
    neg_unsat = unsat(cnf);
  }
  int return_val;
  if (neg_unsat) {
    cout << (check_adder ?
             "l'additionneur est correct" :
             "la formule est valide")
         << endl;
    return_val = EXIT_SUCCESS;
  } else {
    cout << (check_adder ?
             "l'additionneur est incorrect" :
             "la formule n'est pas valide")
         << endl;
    return_val = EXIT_FAILURE;
  }
  free_all(to_check);
  return return_val;
}
